/* joueur.c - Un Joueur */

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <ncurses.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "myfunctions.h"
#include "joueur.h"
#include "circuit_struct.h"
#include "joueur_struct.h"
#include "file_struct.h"
#include "protocole.h"
#include "my_ncurses.h"
#include "rebours.h"

char buf[TAILLE_BUFFER+1];


WINDOW* win_circuit;	  	/* Fenetre du circuit           */
WINDOW* win_info1;		/* Fenetre d'informations       */
WINDOW* win_info2;		/* Fenetre d'informations       */

int tube_fils_pere[2];	  	/* Pour comm. entre les 2 proc  */
int tube_pere_fils[2];	  	/* Pour comm. entre les 2 proc  */

int fd_maitre;	 	  	/* Descripteur tube maitre      */
struct joueur le_j;	  	/* Le joueur	            	*/

int main(int argc, char** argv)
{
	pid_t pid_maitre;	  	/* Pid du maitre            */
	ssize_t n;		  	/* Nombre de caracteres lus */
	int i,j;

	/******************************************
	 * CREATION DES TUBES POUR COMMUNICATION  *
	 * BIDIRECTIONNELLE ENTRE LES 2 PROCESSUS *
	 ******************************************/
	pipe(tube_fils_pere);
	pipe(tube_pere_fils);

	/******************************
	 * VERIFICATION DES ARGUMENTS *
	 ******************************/
	if(argc!=2)
	{
		fprintf(stderr,"usage : %s pid\n",argv[0]);
		return ERROR_ARGS;
	}

	if((pid_maitre=atoi(argv[1]))==0)
		sortie_erreur("pid doit etre un entier\n",ERROR_INT);

	/***************************************************
	 * OUVERTURE TUBE MAITRE - CREATION DU TUBE JOUEUR *
         ***************************************************/
	sprintf(buf,PATTERN_MAITRE,pid_maitre);
	if((fd_maitre=ouvrir_fifo(buf, O_WRONLY))==-1)
		sortie_erreur("Erreur ouverture du tube maitre\n",ERROR_OPEN_FD);

	sprintf(buf,PATTERN_JOUEUR,getpid());
	if(creation_fifo(buf, 0772)==-1)
		sortie_erreur("Erreur creation du tube joueur\n",ERROR_CREAT_FIFO);
	else
		if((le_j.fd_joueur=ouvrir_fifo(buf, O_RDWR))==-1)
			sortie_erreur("Erreur ouverture du tube joueur\n",ERROR_OPEN_FD);

	/*************************
	 * PROCESSUS D'AFFICHAGE *
         *************************/
	switch(le_j.pid_aff=fork())
	{
		case -1 : sortie_erreur("erreur fork\n",ERROR_FORK);
		case  0 : fils(tube_fils_pere,tube_pere_fils); exit(NO_ERROR);
	}

	/*********************************************
	 * SIGNAL QUI PERMET DE QUITTER DANS LE FILS *
	 *********************************************/
	signal(SIGUSR1,quitter);
	signal(SIGINT,SIG_IGN);

	/******************
	 * PROCESSUS PERE *
         ******************/

	close(tube_fils_pere[1]); /* Fermeture en ecriture */
	close(tube_pere_fils[0]); /* Fermeture en lecture  */

	/* Ici le pere attend que le tube du fils soit creer avant de continuer */
	my_read(tube_fils_pere[0],buf,TAILLE_PAQUET);

	/* Ecriture des pid du processus pere et fils dans le tube maitre */
	sprintf(buf,P_CONNEXION_JOUEUR,getpid(),le_j.pid_aff);
	my_write(fd_maitre,buf,TAILLE_PAQUET);

	/* Reception du numero de joueur */
	if((n=my_read(le_j.fd_joueur,buf,TAILLE_PAQUET))>0)
		sscanf(buf,P_NUM_JOUEUR,&le_j.num);

	/* Reception du compte a rebours */
	while((n=my_read(le_j.fd_joueur,buf,TAILLE_PAQUET))>0 && strcmp(buf,P_FIN_COMPTE_REBOURS)!=0)
		my_write(tube_pere_fils[1],buf,TAILLE_PAQUET);

	/* Previent le fils qu'il peut afficher le circuit */
	my_write(tube_pere_fils[1],P_FIN_COMPTE_REBOURS,TAILLE_PAQUET);

	/* Attente des deplacements du joueur */
	while((j=getchar()))
	{
		i=get_accel(j);	
		if(i==-2) break;
		
		/* Envoie du deplacement */
		if(i!=-1)
		{
			sprintf(buf, P_COUP_JOUEUR, le_j.num, i);
			my_write(fd_maitre,buf,TAILLE_PAQUET);
		}
	}

	/* Fermeture des tubes */
	close(fd_maitre);
	close(le_j.fd_joueur);
	close(tube_fils_pere[0]);
	close(tube_pere_fils[1]);

	/* Suppression du tube joueur */
	sprintf(buf,PATTERN_JOUEUR,getpid());
	unlink(buf);

	return NO_ERROR;
}

/* Affichage du circuit stocke dans
le circuit tel_circuit
 */
void affiche_circuit(struct circuit* tel_circuit, WINDOW* win)
{
	int i,j;

	if(wmove(win,0,0)==ERR)
		sortie_erreur_fils("pb ncurses\n",ERROR_NCURSES);

	for(i=0;i<tel_circuit->haut;i++)
		for(j=0;j<tel_circuit->larg;j++)
			if(wprintw(win,"%c",tel_circuit->trace[i][j])==ERR)
				sortie_erreur_fils("pb ncurses\n",ERROR_NCURSES);
}

/* Reception du circuit que l'on stocke dans la
 * structure tel_circuit
 */
void reception_circuit(int tube_aff, struct circuit* tel_circuit)
{
	char* buf2;
	int i,nbCar;

	buf2=NULL;

	/* Reception de la taille du circuit de la forme P_TAILLE_CIRCUIT (protocole.h) */
	my_read_fils(tube_aff,buf,TAILLE_PAQUET);
	sscanf(buf,P_TAILLE_CIRCUIT,&tel_circuit->larg, &tel_circuit->haut);

	/* Allocation du tableau a double dimension pour le circuit */
	tel_circuit->trace=(char**)my_malloc(tel_circuit->larg*tel_circuit->haut*sizeof(char));
	for(i=0;i<tel_circuit->haut;i++)
		tel_circuit->trace[i]=(char*)my_malloc(tel_circuit->larg*sizeof(char));

	/* Allocation d'un buffer temporaire */
	buf2=(char*)my_malloc(tel_circuit->larg*tel_circuit->haut*sizeof(char));

	/* Reception du circuit */
	nbCar=0;
	while(my_read_fils(tube_aff,buf,TAILLE_PAQUET)>0 && strcmp(buf,P_FIN_CIRCUIT)!=0) 
	{	
		strcpy(&buf2[nbCar],buf);
		nbCar+=strlen(buf);	
	}

	/* On ecrit le circuit dans le tableau a double dimensions */
	for(i=0;i<tel_circuit->larg*tel_circuit->haut;i++)
		tel_circuit->trace[(int)i/tel_circuit->larg][i%tel_circuit->larg]=buf2[i];

	/* Desallocation */
	my_free((void*)buf2);
}

/* Bouge le joueur sur le circuit */
void move_joueur(int lig_lu, int col_lu, struct file_struct* telle_file, struct circuit* tel_circuit,int bold, WINDOW* win)
{
	struct coord coortemp;

	if(telle_file->coup>=AFF_MAX)
	{
		/* On recupere la derniere position a effacer - On affiche le car d'origine */
		defiler(telle_file,&coortemp);
		ncurses_print_char(win,tel_circuit->trace[coortemp.x][coortemp.y],coortemp.x,coortemp.y);
	}
	
	/* Sauvegarde la nouvelle position */
	enfiler(telle_file,lig_lu,col_lu);

	/* Gras ou pas */
	if(bold==1)	ncurses_print_bold(win,telle_file->num,lig_lu,col_lu); 
	else		ncurses_print(win,telle_file->num,lig_lu,col_lu); 

	/* Rafraichissement */
	ncurses_refresh(1,win);
}

/* Write avec sortie sur erreur
 * Retourne -> le nombre de d'octets ecrits
 * Sort	    -> en cas d'erreur d'ecriture
 */
ssize_t my_write_fils(int fd, const void *buf, size_t count)
{
	ssize_t ret;

	if((ret=write(fd,buf,count))==-1)
		sortie_erreur_fils("Erreur ecriture\n",ERROR_WRITE_FD);

	return ret;
}

/* Read avec sortie sur erreur
 * Retourne -> le nombre de d'octets lus
 * Sort	    -> en cas d'erreur de lecture
 */
ssize_t my_read_fils(int fd, void *buf, size_t count)
{
	ssize_t ret;

	if((ret=read(fd,buf,count))==-1)
		sortie_erreur_fils("Erreur lecture\n",ERROR_READ_FD);

	return ret;
}

/* Quitte le pere a la reception d'un signal */
void quitter(int signo)
{
	sleep(1);

	/* Fermeture des tubes */
	close(fd_maitre);
	close(le_j.fd_joueur);
	close(tube_fils_pere[0]);
	close(tube_pere_fils[1]);

	/* Suppression du tube joueur */
	sprintf(buf,PATTERN_JOUEUR,getpid());
	unlink(buf);


	exit(NO_ERROR);
}

/* Quitte le fils et envoie un signal au pere */
void sortie_erreur_fils(char* chaine, int errorCode)
{
	kill(getppid(),SIGUSR1);
	sortie_erreur(chaine,errorCode);
}

/* Retourne l'acceleration correspondant a la
 * touche passe en parametre
 */
int get_accel(int touche)
{
	switch(touche)
	{
		case KEY_HAUT: 		return  8;
		case KEY_BAS:  		return  2;
		case KEY_GAUCHE:	return  4;
		case KEY_DROITE:	return  6;
		case KEY_H_GAUCHE:	return  7;
		case KEY_H_DROITE:	return  9;
		case KEY_B_GAUCHE:	return  1;
		case KEY_B_DROITE:	return  3;
		case KEY_NO_CHANGE:	return  5;
		/*case KEY_QUITTER:	return -2;*/
		default:		return -1;
	}
}

/* Chez le fils - processus d'affichage */
void fils(int tube_fils_pere[],int tube_pere_fils[])
{
	struct file_struct* laFileJ;	/* Les 4 dernieres pos. des joueurs */
	struct circuit le_circuit; 	/* Le circuit		            */
	struct joueur le_j;	  	/* Le joueur	                    */
	int joueur_lu;		   	/* Numero du joueur lu              */
	int lig_lu;		  	/* Ligne lue			    */
	int col_lu;		  	/* Colonne lue		            */
	int tour_cur, tour_tot;

	int i,j,nbJoueurs;

	signal(SIGINT, quitte_ncurses);

	close(tube_fils_pere[0]); /* fermeture en lecture  */
	close(tube_pere_fils[1]); /* fermeture en ecriture */

	/******************
	 * TUBE AFFICHAGE *
         ******************/
	sprintf(buf,PATTERN_AFFICHE,getpid());
	if(creation_fifo(buf, 0772)==-1)
		sortie_erreur_fils("Erreur creation du tube\n",ERROR_CREAT_FIFO);
	else
		if((le_j.fd_aff=ouvrir_fifo(buf, O_RDWR))==-1)
			sortie_erreur_fils("Erreur ouverture du tube affichage\n",ERROR_OPEN_FD);
					
	/* On previent le pere qu'il peut envoyer les pids */
	/* (et surtout que le tube d'affichage est pret !) */
	sprintf(buf,"continue");
	my_write_fils(tube_fils_pere[1],buf,TAILLE_PAQUET);

	/************************
	 * RECEPTION DU CIRCUIT *
	 ************************/
	reception_circuit(le_j.fd_aff,&le_circuit);

	/*********************************
	 * INITIALISATION DE L'AFFICHAGE *
	 *********************************/
	ncurses_init(le_circuit.haut+1,le_circuit.larg+2+2*LARG_CARAC);
	win_circuit=ncurses_wcreate(le_circuit.haut+1,le_circuit.larg,0,0);
	win_info1=ncurses_wcreate(le_circuit.haut,LARG_CARAC,0,le_circuit.larg+1);
	win_info2=ncurses_wcreate(le_circuit.haut,LARG_CARAC,0,le_circuit.larg+LARG_CARAC+2);

	/************************
	 * AFFICHAGE DU CIRCUIT *
	 ************************/
	affiche_circuit(&le_circuit,win_circuit);
	ncurses_refresh(3,win_circuit,win_info1,win_info2);

	/***************************************************
	 * RECEPTION DE SA POSITION SUR LA LIGNE DE DEPART *
         ***************************************************/
	my_read(le_j.fd_aff,buf,TAILLE_PAQUET);
	sscanf(buf,P_POS_JOUEUR,&le_j.num, &le_j.lig_cur, &le_j.col_cur);

	/* Ecriture du numero de joueur sur le circuit */
	ncurses_print_bold(win_circuit,le_j.num,le_j.lig_cur,le_j.col_cur);
	ncurses_refresh(1,win_circuit);

	/* Allocation - Initialisation - Enfile 1ere pos joueur */
	laFileJ=(struct file_struct*)my_malloc(sizeof(struct file_struct));
	init_file_struct(&laFileJ[0],le_j.num);
	enfiler(&laFileJ[0],le_j.lig_cur,le_j.col_cur);

	/**********************************************
	 * RECEPTION DES POSITIONS DES AUTRES JOUEURS *
         **********************************************/
	nbJoueurs=1;
	while(my_read(le_j.fd_aff,buf,TAILLE_PAQUET)>0 && strcmp(buf,P_FIN_POS_AUTRES_JOUEURS)!=0)
	{
		/* Enregistrement */
		sscanf(buf,P_POS_AUTRES_JOUEURS,&joueur_lu,&lig_lu,&col_lu);

		/* Reallocation - Initialisation pour ce genre - Enfile position */
		laFileJ=(struct file_struct*)my_realloc(laFileJ,sizeof(struct file_struct)*(nbJoueurs+1));
		init_file_struct(&laFileJ[nbJoueurs],joueur_lu);
		enfiler(&laFileJ[nbJoueurs],lig_lu,col_lu);

		/* Ecriture du numero de joueur sur le circuit */
		ncurses_print(win_circuit, joueur_lu, lig_lu, col_lu);
		ncurses_refresh(1,win_circuit);

		nbJoueurs++;
	}

	/***********************************************
	 * ATTENTE DU SIGNAL DE DEMARRAGE DE LA COURSE *
	 ***********************************************/
	i=1;
	while(my_read(tube_pere_fils[0],buf,TAILLE_PAQUET)>0 && strcmp(buf,P_FIN_COMPTE_REBOURS)!=0)
	{
		sscanf(buf,P_COMPTE_REBOURS,&j);

		/* On affiche go a la place du 0 */
		if(j==0)
		{
			/* Go !*/
			ncurses_print_string(win_info1, donne_carac(G), 0, 0);
			ncurses_print_string(win_info2, donne_carac(O), 0, 0);
			ncurses_refresh(2, win_info1, win_info2);
			continue;
		}

		ncurses_print_string(win_info1,donne_carac(j/10),0,0);
		ncurses_print_string(win_info2,donne_carac(j%10),0,0);
		ncurses_refresh(2,win_info1,win_info2);
	}

	/**********************************
	 * ATTENTE RECEPTION DE POSITIONS *
	 **********************************/
	while(my_read(le_j.fd_aff,buf,TAILLE_PAQUET)>0 && strcmp(buf,P_FIN_POS_JOUEUR)!=0)
	{
		/* Verifie si le joueur a gagne */
		if(strcmp(buf,P_GAGNE)==0)
		{
			ncurses_print_string(win_info1, "Gagne !!!", 0,0);
			ncurses_refresh(2, win_info1, win_info2);
			continue;
		}

		/* Position du joueur a l'arrivee */
		if(strncmp(buf,P_ARR,2)==0)
		{
			sscanf(buf,P_ARR, &i, &j);
			sprintf(buf,"Arr : %d/%d",i,j);
			ncurses_print_string(win_info1, buf, 0,0);
			ncurses_refresh(2, win_info1, win_info2);
			continue;
		}

		/* Reception d'infos sur les tours */
		if(strncmp(buf,P_TOUR,2)==0)
		{
			sscanf(buf,P_TOUR, &joueur_lu, &tour_cur, &tour_tot);
			sprintf(buf,"Tour %d/%d\n", tour_cur, tour_tot);

			ncurses_clear(2, win_info1, win_info2);
			ncurses_print_string(win_info1, buf, 0,0);
			ncurses_refresh(2, win_info1, win_info2);
			continue;
		}

		/* Analyse de la chaine */
		sscanf(buf,P_NOUV_POS_JOUEUR,&joueur_lu, &lig_lu, &col_lu);

		/* Affichage position si c'est le joueur courant */
		if(le_j.num==joueur_lu)
		{
			sprintf(buf,"(%2d ; %2d)     \n", lig_lu, col_lu);
			ncurses_print_string(win_info1, buf, 1,0);
			ncurses_refresh(2, win_info1, win_info2);
		}

		/* Cherche le joueur dans les files */
		for(i=0;i<nbJoueurs;i++)
			if(laFileJ[i].num==joueur_lu)
				break;

		/* Deplacement du joueur */
		move_joueur(lig_lu, col_lu, &laFileJ[i], &le_circuit, le_j.num==joueur_lu, win_circuit);
	}

	/*************
	 * NETTOYAGE *
	 *************/
	ncurses_end(3,win_circuit, win_info1, win_info2);

	/* Fermeture des tubes */
	close(le_j.fd_aff);
	close(tube_fils_pere[0]);
	close(tube_pere_fils[0]);

	/* Suppression du tube affichage */
	sprintf(buf,PATTERN_AFFICHE,getpid());
	unlink(buf);

	/* Free */
	for(i=0;i<le_circuit.haut;i++)
		my_free(le_circuit.trace[i]);
	my_free(le_circuit.trace);
	my_free(laFileJ);

	/* Previent le pere */
	kill(getppid(),SIGUSR1);

	exit(NO_ERROR);		

}

void quitte_ncurses(int signo)
{
	ncurses_end(3,win_circuit, win_info1, win_info2);

	kill(getppid(),SIGUSR1);
	exit(NO_ERROR);
}
